#include "RegionStats.h"
#include "Error.h"
#include "CigarRoller.h"

RegionStats::RegionStats(){ Init(); }

RegionStats::RegionStats(String &input)
{
  Init();
  LoadRegionList(input);
}
RegionStats::~RegionStats() {}

void RegionStats::Init()
{
  minOverlapLen = 1; 
  minMapQuality = 0;
  nReads = 0;
  nTotalMappedReads = 0;
  nMapped2Targets = 0;
  norm_by_mapped2target = false;
  uniq = false;
}

void RegionStats::SetUniqFlag(bool flag) { uniq = flag; }

void RegionStats::CalcRegionStats(StringArray &bamFiles)
{
 for(int i=0; i<bamFiles.Length(); i++) 
 {
   if(CalcRegionStats(bamFiles[i]))
   	error("Calculating stats for bam file %s failed!\n", bamFiles[i].c_str());
 }
}

int RegionStats::CalcRegionStats(String &bamFile)
{
  if(norm_by_mapped2target==false) {
   printf("Counting number of mapped reads in the bam file ...\n");
   CountMappedReads(bamFile);
 }

  String bamFileIndex = bamFile+".bai";
  SamFile samIn;

  if(!samIn.OpenForRead(bamFile.c_str()))
    {
      fprintf(stderr, "%s\n", samIn.GetStatusMessage());
      return(samIn.GetStatus());
    }

  if(!samIn.ReadBamIndex(bamFileIndex.c_str()))
  {
      fprintf(stderr, "%s\n", samIn.GetStatusMessage());
      return(samIn.GetStatus());
  }

  SamFileHeader samHeader;
  SamRecord samRecord;

  if(!samIn.ReadHeader(samHeader))
    {
     fprintf(stderr, "help%s\n", samIn.GetStatusMessage());
     return(samIn.GetStatus());
    }

  int numRegionRecords = 0;  
  int start, end, gstart, gend, atCnt, gcCnt;
  String refLabel, CSE; //Chr_Start_End
  double gc;
  StringArray tokens;
  std::map<String, int> geneReadCount;
  CigarRoller cigarRoller;

  // contigs are contig+"|"+gene so that # of contigs eq # of genes
  for(unsigned int i=0; i<contigs.size(); i++)
    {
      tokens.ReplaceTokens(contigs[i], "|");
      refLabel = tokens[0];
      for(unsigned int j=0; j<genomeRegions[contigs[i]].size(); j++)
	{

	  start = genomeRegions[contigs[i]][j].first;
	  end   = genomeRegions[contigs[i]][j].second;
	  CSE = contigs[i]+":"+start+":"+end;
	  numRegionRecords = 0;

	  //if(start+(minOverlapLen-1)<= end-(minOverlapLen-1))
	  if(start>end) error("The start position %d is larger the the end position %d\n", start, end);

         if(--geneStats[contigs[i]].segCount==0) 
  	 geneReadCount.clear();

	 //The input is 0-based start and Mary Kate's library uses 1-based. So add 1 to the start
	  // if(start+(minOverlapLen-1)+1< end-(minOverlapLen-1) ) 
	  //  samIn.SetReadSection(refLabel.c_str(), start+(minOverlapLen-1)+1, end-(minOverlapLen-1));
	  //else
	    samIn.SetReadSection(refLabel.c_str(), start, end);	  

	  bool isMatch = false;
	  
	  while(samIn.ReadRecord(samHeader, samRecord))
	    { 
	      if(samRecord.getMapQuality()<minMapQuality) continue; 
	      // Ignore a read if this read has been counted to aother exon in the same gene
	      if(++geneReadCount[samRecord.getReadName()]>1 && uniq) { continue;}
	      
	      cigarRoller.Set(samRecord.getCigar());
		   int refOffset;
	      isMatch = false;

	      // Check if the cigar has 'M' bases in the region

		  refOffset = start - samRecord.get1BasedPosition();

	     //If refOffset < 0, it means that the region can not be completely coverged by all 'N's
	     if(refOffset<=0) {
		 if(end-samRecord.get0BasedPosition()+1 >= minOverlapLen) isMatch=true;
		 else continue;
		}
	     else //Check whether this region is spanned by enough M's
	     {
		int overlapCnt = 0;
		for(int k=refOffset; k<=end-start+refOffset; k++)
		  {
		  	//if(cigarRoller.getQueryIndex(k)!=CigarRoller::INDEX_NA) { isMatch = true; break;}
		  	if(cigarRoller.getQueryIndex(k)!=CigarRoller::INDEX_NA) { overlapCnt++; if(overlapCnt>=minOverlapLen) {isMatch=true; break; } }
		  }
              }

		  if(isMatch==true) numRegionRecords++;

	    } //End of while loop
	  if(--geneStats[contigs[i]].segCount==0) geneReadCount.clear();
	  
	  geneStats[contigs[i]].nReads += numRegionRecords; //gene level
	  genomeRegionStats[CSE].nReads = numRegionRecords; //exon level
	  nMapped2Targets += numRegionRecords;
	  
	  //// Calculate GC content
	  //exon level
	  atCnt = 0; gcCnt=0;
	  gstart = referencegenome.getGenomePosition(refLabel.c_str(), start);
	  gend = gstart + (end-start);
	  gc = CalcRegionGCContent(referencegenome,gstart, gend, atCnt, gcCnt);
	  genomeRegionStats[CSE].gcContent = gc;
	  genomeRegionStats[CSE].atCnt = atCnt;
	  genomeRegionStats[CSE].gcCnt = gcCnt;
	  regionReadCountInGCBins[toGCBin(gc)] += numRegionRecords;
	  
	  // gene level
	  geneStats[contigs[i]].atCnt += atCnt;
	  geneStats[contigs[i]].gcCnt += gcCnt;
	}
    }
  return(0);
}

void RegionStats::LoadReferenceGenome(String &reference)
{
  bool memoryMap = false;
  fprintf(stderr,"Loading reference... ");  
  referencegenome.setReferenceName(reference.c_str());
  referencegenome.useMemoryMap(memoryMap);
  if (referencegenome.open())
  {
    fprintf(stderr, "Failed to open reference index and is creating one...\n");
    if(referencegenome.create(false))
    	error("Creating reference index failed!\n");
    if(referencegenome.open())
    	error("Open newly created reference failed!\n");
   }
  fprintf(stderr, "DONE! Total sequence length %u\n", referencegenome.sequenceLength());
}


void RegionStats::LoadRegionList(String &inputList)
{
  printf("%s\n","Loading region list...");
  FILE *in = fopen(inputList.c_str(), "r");
  if(in==NULL) error("Open region input file %s failed!\n", inputList.c_str());
  StringArray tokens;
  String buffer;
  while(!feof(in))
    {
      buffer.ReadLine(in);
      if (buffer.IsEmpty() || buffer[0] == '#') continue;
      tokens.ReplaceTokens(buffer);
      if(tokens.Length()<3)
	error("Too few columns (minimum 3 required): %s\n", buffer.c_str());

      std::pair<int, int> start_end;
      start_end.first = tokens[1].AsInteger();
      start_end.second = tokens[2].AsInteger();
      if(start_end.first>=start_end.second) // positions are 0-based. Otherwise == is valid
      	error("Region end is equal or smaller than the start: %s!\n", buffer.c_str());      
      //genomeRegions_lines[tokens[0]].push_back(buffer);
      //genomeRegions[tokens[0]].push_back(start_end);
      
      String geneKey;
      if(tokens.Length()==3) geneKey = tokens[0] + "|" + tokens[0] + ":" + tokens[1] + "-" + tokens[2];
      else  geneKey = tokens[0]+"|"+tokens[3];

      genomeRegions_lines[geneKey].push_back(buffer);
      genomeRegions[geneKey].push_back(start_end);
      geneStats[geneKey].segCount++;
      geneStats[geneKey].totalLen += (start_end.second - start_end.first);
    }

  fclose(in);

  // Chromosome info
  contigs.clear();
  std::map<String, vector<std::pair<int, int> > >::iterator p;
  for(p=genomeRegions.begin(); p!=genomeRegions.end(); p++)
    {
      contigs.push_back(p->first);
      /*
      for(unsigned int i=1; i<genomeRegions[p->first].size(); i++)
  	   if(genomeRegions[p->first][i].first<genomeRegions[p->first][i-1].first)
	     error("Input coordinates are not in order: %s %d %d!\n", p->first.c_str(),genomeRegions[p->first][i].first,genomeRegions[p->first][i].second);
	  */
    }
}


void RegionStats::Reset()
{
}

void RegionStats::CalcGeneGCContent()
{
 for(unsigned int i=0; i<contigs.size(); i++)
   {
		double gc = double(geneStats[contigs[i]].gcCnt)/(geneStats[contigs[i]].atCnt+geneStats[contigs[i]].gcCnt)*100;
		geneStats[contigs[i]].gcContent = gc;
		geneReadCountInGCBins[toGCBin(gc)] += geneStats[contigs[i]].nReads;
	}
}

void RegionStats::CalcRegionGCContent()
{
  double gc;
  int start, end;
  genomeIndex_t gstart, gend;
  String CSE; //Chr_Start_End
  for(unsigned int i=0; i<contigs.size(); i++)
    {
      for(unsigned int j=0; j<genomeRegions[contigs[i]].size(); j++)
	{
	  start = genomeRegions[contigs[i]][j].first;
	  end   = genomeRegions[contigs[i]][j].second;
	  CSE = contigs[i]+":"+start+":"+end;
	  gstart = referencegenome.getGenomePosition(contigs[i].c_str(), start);
	  gend = gstart + (end-start);
	  int atCnt, gcCnt;
	  gc = CalcRegionGCContent(referencegenome,gstart, gend, atCnt, gcCnt);
	  genomeRegionStats[CSE].gcContent = gc;
	  genomeRegionStats[CSE].atCnt = atCnt;
	  genomeRegionStats[CSE].gcCnt = gcCnt;
	}
    }
}

double RegionStats::CalcRegionGCContent(GenomeSequence &ref, uint32_t start, uint32_t end, int &atCnt, int &gcCnt)
{
  atCnt=0;
  gcCnt=0;
  for(uint32_t i=start; i<end; i++)
    {
      if(toupper(ref[i])=='A' || toupper(ref[i])=='T') atCnt++;
      if(toupper(ref[i])=='C' || toupper(ref[i])=='G') gcCnt++;
    }
  double gcContent = 100*double(gcCnt)/(atCnt+gcCnt);
  return(gcContent);
}

void RegionStats::CalcRegionReadCountInGCBins()
{
  if(referencegenome.sequenceLength()==0) return;
  double gc;
  int start, end;

  String CSE; //Chr_Start_End
  for(unsigned int i=0; i<contigs.size(); i++)
    {
      for(unsigned int j=0; j<genomeRegions[contigs[i]].size(); j++)
	{
	  start = genomeRegions[contigs[i]][j].first;
	  end   = genomeRegions[contigs[i]][j].second;
	  CSE = contigs[i]+":"+start+":"+end;
	  gc = genomeRegionStats[CSE].gcContent;
	  regionReadCountInGCBins[toGCBin(gc)] += genomeRegionStats[CSE].nReads;
	}
    }
}

int RegionStats::toGCBin(double gc)
{
 return(floor(gc+0.5));
}


void RegionStats::OutputRegionStats(String &outFile)
{
  FILE *fh;
  if(outFile.Length()==0) fh = stdout;
  else fh = fopen(outFile.c_str(), "w");
  if(fh==NULL)
    error("Open out file %s failed!\n", outFile.c_str());

  int start, end;
  fprintf(fh, "##Mapped2targets reads: %lu", nMapped2Targets);
  if(norm_by_mapped2target==false) fprintf(fh, "\tTotal mapped reads: %lu", nTotalMappedReads);
  fprintf(fh,"\n#Chr\tStart\tEnd\tGene");
  fprintf(fh, "\tLen\tReadCnt\tRPM\tRPKM");
  fprintf(fh, "\t%%GC\tnReadGCBin\n");

  for(unsigned int i=0; i<contigs.size(); i++)
    {
      for(unsigned int j=0; j<genomeRegions[contigs[i]].size(); j++)
	{
	  start = genomeRegions[contigs[i]][j].first;
	  end   = genomeRegions[contigs[i]][j].second;
	  String CSE = contigs[i]+":"+start+":"+end;
	  int regionLen = end - start; // poistions are 0-based

	  double RPM,RPKM;

 	  if(norm_by_mapped2target) {
	    RPM = double(genomeRegionStats[CSE].nReads)/(double(nMapped2Targets)/1000000);
	    RPKM = RPM/regionLen*1000;
          }
          else {
	    RPM = double(genomeRegionStats[CSE].nReads)/(double(nTotalMappedReads)/1000000);
	    RPKM = RPM/regionLen*1000;
	  }

	  fprintf(fh, "%s\t%d\t%lu\t%.3f\t%.3f", genomeRegions_lines[contigs[i]][j].c_str(), end-start, genomeRegionStats[CSE].nReads, RPM, RPKM);
	  fprintf(fh, "\t%.2f\t%lu", genomeRegionStats[CSE].gcContent, regionReadCountInGCBins[toGCBin(genomeRegionStats[CSE].gcContent)]);
	   fprintf(fh, "\n");
	}
    }
  fclose(fh);
}


void RegionStats::OutputGeneStats(String &outFile)
{
  FILE *fh;
  if(outFile.Length()==0) fh = stdout;
  else fh = fopen(outFile.c_str(), "w");
  if(fh==NULL)
    error("Open out file %s failed!\n", outFile.c_str());

  fprintf(fh, "##Mapped2targets reads: %lu", nMapped2Targets);
  if(norm_by_mapped2target==false) fprintf(fh, "\tTotal mapped reads: %lu", nTotalMappedReads);
  fprintf(fh, "\n#Gene\tLen\tReadCnt\tRPM\tRPKM");
  fprintf(fh, "\t%%GC\tnReadGCBin\n");

  for(unsigned int i=0; i<contigs.size(); i++)
    {
	  double RPM, RPKM;

	  if(norm_by_mapped2target) {
	   RPM = double(geneStats[contigs[i]].nReads)/(double(nMapped2Targets)/1000000);
	   RPKM = RPM/geneStats[contigs[i]].totalLen*1000;
	  }
	  else {
	   RPM = double(geneStats[contigs[i]].nReads)/(double(nTotalMappedReads)/1000000);
	   RPKM = RPM/geneStats[contigs[i]].totalLen*1000;
          }

     StringArray tokens;
     tokens.ReplaceTokens(contigs[i], "|");
	  fprintf(fh, "%s\t%d\t%lu\t%.3f\t%.3f", tokens[1].c_str(), geneStats[contigs[i]].totalLen, geneStats[contigs[i]].nReads, RPM, RPKM);
	  fprintf(fh, "\t%.2f\t%lu", geneStats[contigs[i]].gcContent, geneReadCountInGCBins[toGCBin(geneStats[contigs[i]].gcContent)]);
	   fprintf(fh, "\n");
	}
  fclose(fh);
}

int RegionStats::CountMappedReads(String &bamFile)
{
  String bamFileIndex = bamFile+".bai";
  SamFile samIn;

  if(!samIn.OpenForRead(bamFile.c_str()))
    {
      fprintf(stderr, "%s\n", samIn.GetStatusMessage());
      return(samIn.GetStatus());
    }

  if(!samIn.ReadBamIndex(bamFileIndex.c_str()))
  {
      fprintf(stderr, "%s\n", samIn.GetStatusMessage());
      return(samIn.GetStatus());
  }

  SamFileHeader samHeader;
  SamRecord samRecord;

  if(!samIn.ReadHeader(samHeader))
    {
     fprintf(stderr, "help%s\n", samIn.GetStatusMessage());
     return(samIn.GetStatus());
    }

   nTotalMappedReads = 0;
   while(samIn.ReadRecord(samHeader, samRecord))
    { 
     if(samRecord.getMapQuality()<minMapQuality) continue; 
     nTotalMappedReads++;
    }
   return(0);
}
